# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import timezone
import unittest
from unittest.mock import patch

from swh.scheduler import utils


class UtilsTest(unittest.TestCase):
    @patch("swh.scheduler.utils.datetime")
    def test_create_oneshot_task_dict_simple(self, mock_datetime):
        mock_datetime.now.return_value = "some-date"

        actual_task = utils.create_oneshot_task_dict("some-task-type")

        expected_task = {
            "policy": "oneshot",
            "type": "some-task-type",
            "next_run": "some-date",
            "arguments": {"args": [], "kwargs": {},},
        }

        self.assertEqual(actual_task, expected_task)
        mock_datetime.now.assert_called_once_with(tz=timezone.utc)

    @patch("swh.scheduler.utils.datetime")
    def test_create_oneshot_task_dict_other_call(self, mock_datetime):
        mock_datetime.now.return_value = "some-other-date"

        actual_task = utils.create_oneshot_task_dict(
            "some-task-type", "arg0", "arg1", priority="high", other_stuff="normal"
        )

        expected_task = {
            "policy": "oneshot",
            "type": "some-task-type",
            "next_run": "some-other-date",
            "arguments": {
                "args": ("arg0", "arg1"),
                "kwargs": {"other_stuff": "normal"},
            },
            "priority": "high",
        }

        self.assertEqual(actual_task, expected_task)
        mock_datetime.now.assert_called_once_with(tz=timezone.utc)

    @patch("swh.scheduler.utils.datetime")
    def test_create_task_dict(self, mock_datetime):
        mock_datetime.now.return_value = "date"

        actual_task = utils.create_task_dict(
            "task-type",
            "recurring",
            "arg0",
            "arg1",
            priority="low",
            other_stuff="normal",
            retries_left=3,
        )

        expected_task = {
            "policy": "recurring",
            "type": "task-type",
            "next_run": "date",
            "arguments": {
                "args": ("arg0", "arg1"),
                "kwargs": {"other_stuff": "normal"},
            },
            "priority": "low",
            "retries_left": 3,
        }

        self.assertEqual(actual_task, expected_task)
        mock_datetime.now.assert_called_once_with(tz=timezone.utc)



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import logging
import sys
import time
from typing import Any, Dict, Optional

import click

from . import cli

logger = logging.getLogger(__name__)


def destination_from_pattern(ctx: click.Context, pattern: Optional[str]):
    """Get the celery destination pattern from host and type values"""
    if pattern is None:
        logger.debug("Matching all workers")
    elif "*" in pattern:
        ctx.obj["inspect"].pattern = pattern
        ctx.obj["inspect"].matcher = "glob"
        logger.debug("Using glob pattern %s", pattern)
    else:
        destination = pattern.split(",")
        ctx.obj["inspect"].destination = destination
        logger.debug("Using destinations %s", ", ".join(destination))


@cli.group("celery-monitor")
@click.option(
    "--timeout", type=float, default=3.0, help="Timeout for celery remote control"
)
@click.option("--pattern", help="Celery destination pattern", default=None)
@click.pass_context
def celery_monitor(ctx: click.Context, timeout: float, pattern: Optional[str]) -> None:
    """Monitoring of Celery"""
    from swh.scheduler.celery_backend.config import app

    ctx.obj["timeout"] = timeout
    ctx.obj["inspect"] = app.control.inspect(timeout=timeout)

    destination_from_pattern(ctx, pattern)


@celery_monitor.command("ping-workers")
@click.pass_context
def ping_workers(ctx: click.Context) -> None:
    """Check which workers respond to the celery remote control"""

    response_times = {}

    def ping_callback(response):
        rtt = time.monotonic() - ping_time
        for destination in response:
            logger.debug("Got ping response from %s: %r", destination, response)
            response_times[destination] = rtt

    ctx.obj["inspect"].callback = ping_callback

    ping_time = time.monotonic()
    ret = ctx.obj["inspect"].ping()

    if not ret:
        logger.info("No response in %f seconds", time.monotonic() - ping_time)
        ctx.exit(1)

    for destination in ret:
        logger.info(
            "Got response from %s in %f seconds",
            destination,
            response_times[destination],
        )

    ctx.exit(0)


@celery_monitor.command("list-running")
@click.option(
    "--format",
    help="Output format",
    default="pretty",
    type=click.Choice(["pretty", "csv"]),
)
@click.pass_context
def list_running(ctx: click.Context, format: str):
    """List running tasks on the lister workers"""
    from ast import literal_eval
    import csv

    response_times = {}

    def active_callback(response):
        rtt = time.monotonic() - active_time
        for destination in response:
            response_times[destination] = rtt

    ctx.obj["inspect"].callback = active_callback

    active_time = time.monotonic()
    ret = ctx.obj["inspect"].active()

    if not ret:
        logger.info("No response in %f seconds", time.monotonic() - active_time)
        ctx.exit(1)

    def pretty_task_arguments(task: Dict[str, Any]) -> str:
        arg_list = []
        for arg in task["args"]:
            arg_list.append(repr(arg))
        for k, v in task["kwargs"].items():
            arg_list.append(f"{k}={v!r}")

        return f'{task["name"]}({", ".join(arg_list)})'

    def get_task_data(worker: str, task: Dict[str, Any]) -> Dict[str, Any]:
        duration = time.time() - task["time_start"]
        return {
            "worker": worker,
            "name": task["name"],
            "args": literal_eval(task["args"]),
            "kwargs": literal_eval(task["kwargs"]),
            "duration": duration,
            "worker_pid": task["worker_pid"],
        }

    if format == "csv":
        writer = csv.DictWriter(
            sys.stdout, ["worker", "name", "args", "kwargs", "duration", "worker_pid"]
        )
        writer.writeheader()

        def output(data: Dict[str, Any]):
            writer.writerow(data)

    elif format == "pretty":

        def output(data: Dict[str, Any]):
            print(
                f"{data['worker']}: {pretty_task_arguments(data)} "
                f"[for {data['duration']:f}s, pid={data['worker_pid']}]"
            )

    else:
        logger.error("Unknown format %s", format)
        ctx.exit(127)

    for worker, active in sorted(ret.items()):
        if not active:
            logger.info("%s: no active tasks", worker)
            continue

        for task in sorted(active, key=lambda t: t["time_start"]):
            output(get_task_data(worker, task))

    ctx.exit(0)



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from flask import url_for
import pytest

from swh.scheduler.api.client import RemoteScheduler
import swh.scheduler.api.server as server
from swh.scheduler.tests.test_scheduler import TestScheduler  # noqa

# tests are executed using imported class (TestScheduler) using overloaded
# swh_scheduler fixture below


# the Flask app used as server in these tests
@pytest.fixture
def app(swh_db_scheduler):
    assert hasattr(server, "scheduler")
    server.scheduler = swh_db_scheduler
    yield server.app


# the RPCClient class used as client used in these tests
@pytest.fixture
def swh_rpc_client_class():
    return RemoteScheduler


@pytest.fixture
def swh_scheduler(swh_rpc_client, app):
    yield swh_rpc_client


def test_site_map(flask_app_client):
    sitemap = flask_app_client.get(url_for("site_map"))
    assert sitemap.headers["Content-Type"] == "application/json"

    rules = set(x["rule"] for x in sitemap.json)
    # we expect at least these rules
    expected_rules = set(
        "/" + rule
        for rule in (
            "lister/get_or_create",
            "lister/update",
            "origins/get",
            "origins/record",
            "priority_ratios/get",
            "task/create",
            "task/delete_archived",
            "task/disable",
            "task/filter_for_archive",
            "task/get",
            "task/grab_ready",
            "task/peek_ready",
            "task/search",
            "task/set_status",
            "task_run/end",
            "task_run/get",
            "task_run/schedule",
            "task_run/schedule_one",
            "task_run/start",
            "task_type/create",
            "task_type/get",
            "task_type/get_all",
        )
    )
    assert rules == expected_rules


def test_root(flask_app_client):
    root = flask_app_client.get("/")
    assert root.status_code == 200
    assert b"Software Heritage scheduler RPC server" in root.data
